#!/usr/bin/python

from twisted.internet import reactor, protocol, error
from autobahn.websocket import WebSocketServerFactory, \
                               WebSocketServerProtocol, \
                               listenWS
from twisted.python.log import startLogging, msg
from twisted.web.server import Site
from twisted.web.static import File
import sys
import subprocess
from datetime import datetime
import flash_latest
startLogging(sys.stdout)


class ProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, consolelog_factory):
        self.consolelog_factory = consolelog_factory
        self.buffer = []
        self.consolelog_factory.broadcast(wrapMessage("STATUS", "running"))

    def outReceived(self, message):
        self.appendMessage(message)

    def errReceived(self, data):
        print "Error: %s" % data

    def processExited(self, reason):
        self.consolelog_factory.process = None
        if reason.check(error.ProcessTerminated):
            self.appendMessage("==== TESTING CANCELLED ====")
        elif reason.check(error.ProcessDone):
            self.appendMessage("==== TESTING COMPLETE ====")
            self.appendMessage("==== REBOOTING DEVICE... ====")
            subprocess.call(["adb", "reboot"])
        self.consolelog_factory.broadcast(wrapMessage("STATUS", "stopped"))

    def appendMessage(self, message):
        self.consolelog_factory.broadcast(msgConsole(message))
        self.buffer.append(message)


PROTOCOL_SEPARATOR = ":::PROTO:::"

def wrapMessage(header, content):
    data = "%s%s%s" % (header, PROTOCOL_SEPARATOR, content)
    return data

def unwrapMessage(data):
    s = data.split(PROTOCOL_SEPARATOR)
    return s[0], s[1]

def msgConsole(message):
    return wrapMessage("CONSOLE", message)


class ConsoleLogProtocol(WebSocketServerProtocol):
    def onOpen(self):
        self.factory.register(self)
        self.sendMessage(wrapMessage("FLASHING", self.factory.flashing))
        self.sendMessage(wrapMessage("PLUGGED", self.factory.plugged))
        if self.factory.process:
            status = "running" if self.factory.process else "stopped"
            self.sendMessage(wrapMessage("STATUS", status))
            self.sendMessage(wrapMessage("TESTS_LAUNCHED", self.factory.url))
            for line in self.factory.process.buffer:
                self.sendMessage(msgConsole(line))

    def connectionLost(self, reason):
        WebSocketServerProtocol.connectionLost(self, reason)
        self.factory.unregister(self)

    def onMessage(self, message, isBinary):
        header, content = unwrapMessage(message)
        msg(header, content)
        if header == "LAUNCH_TESTS":
            self.launchTests(content)
        elif header == "CANCEL":
            self.cancelTests()
        elif header == "REBOOT":
            self.reboot()

    def launchTests(self, debs_url):
        if self.factory.process:
            msg("Another process is already running.")
            return

        if self.factory.flashing:
            msg("Flashing in progress.")
            return

        self.factory.process = ProcessProtocol(self.factory)
        command = "run_tests_with_debs.py"
        FORMAT = '%Y.%m.%d_at_%H:%M:%S'
        logfile = "/home/phablet/%s_tests.log" % (datetime.now().strftime(FORMAT))
        self.factory.url = debs_url
        args = [command, self.factory.url, logfile]
        self.factory.processTransport = reactor.spawnProcess(self.factory.process, command, args, {}, usePTY=True)
        self.factory.broadcast(wrapMessage("TESTS_LAUNCHED", self.factory.url))

    def cancelTests(self):
        self.factory.processTransport.signalProcess("KILL")

    def reboot(self):
        subprocess.call(["adb", "reboot"])


class ConsoleLogProtocolFactory(WebSocketServerFactory):
    protocol = ConsoleLogProtocol
        
    def __init__(self, *args, **kwargs):
        WebSocketServerFactory.__init__(self, *args, **kwargs)
        self.clients = []
        self.process = None
        self.processTransport = None
        self.url = ""
        self.flashing = False
        self.plugged = False

    def register(self, client):
        msg("Registered client %s" % client)
        if not client in self.clients:
            self.clients.append(client)

    def unregister(self, client):
        msg("Unregistered client %s" % client)
        if client in self.clients:
            self.clients.remove(client)

    def broadcast(self, message):
        for client in self.clients:
            client.sendMessage(message)

    def setFlashing(self, flashing):
        self.flashing = flashing
        self.broadcast(wrapMessage("FLASHING", flashing))

    def setPlugged(self, plugged):
        if plugged != self.plugged:
            self.plugged = plugged
            self.broadcast(wrapMessage("PLUGGED", plugged))


def is_device_plugged():
    return subprocess.check_output(["adb", "get-state"]).rstrip() == "device"

if __name__ == "__main__":
    factory = ConsoleLogProtocolFactory("ws://0.0.0.0:9000", debug=False)
    listenWS(factory)

    def flash():
        try:
            flash_latest.flash()
        finally:
            factory.setFlashing(False)
            reactor.callLater(3600, try_flashing)
 
    def try_flashing():
        if factory.process:
            return

        factory.setFlashing(True)
        reactor.callInThread(flash)

    def detect_device():
        plugged = is_device_plugged()
        factory.setPlugged(plugged)
        if plugged:
            reactor.callLater(5, detect_device)
        else:
            reactor.callInThread(wait_for_device)

    def wait_for_device():
        subprocess.call(["adb", "wait-for-device"])
        reactor.callLater(0, detect_device)

    resource = File("site")
    html_factory = Site(resource)
    reactor.listenTCP(8000, html_factory)
    reactor.callLater(0, try_flashing)
    reactor.callLater(0, detect_device)
    reactor.run()

